package de.mpg.ipp.codac.business.schema.testcase;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.jdo.annotations.Element;
import javax.jdo.annotations.FetchGroup;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;

import de.mpg.ipp.codac.business.schema.testcase.W7XDbNamedObject;

@PersistenceCapable
@FetchGroup(name = "deep", members = { @Persistent(name = "users") })
public class TestUserGroup
    extends W7XDbNamedObject {

    static final long serialVersionUID = 0;

    @Persistent(column = "users")
    @Element(types = TestUser.class)
    private List<TestUser> users;

    /**
     * Instantiates a new user group.
     */
    public TestUserGroup() {
        setDefault();
    }

    public void setDefault() {
        super.setDefault();
        users = new ArrayList<>();

    }

    @Override
    public boolean equals( Object obj ) {

        if (this == obj)
            return true;
        if (!(obj instanceof TestUserGroup))
            return false;

        TestUserGroup g = (TestUserGroup) obj;

        /* marcl:140212
         * Even if users and subgroups are equal, we have cases where we want to have separate groups. 
         * In my opinion we have to take the sysName for this special case into account, 
         * because it has the same function as the uid of the User class!
         * TODO: add String groupID to differ from sys-Attribute sysName!
         */
        if (!sysName.equals(g.getSysName()))
            return false;

        if (users == null ^ g.users == null)
            return false;

        if (users != null && g.users != null) {
            if (users.size() != g.users.size())
                return false;
            if (!users.containsAll(g.users))
                return false;
            // because duplicate entries are not forbidden - check the other way, too:
            if (!g.users.containsAll(users))
                return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        int hash = 23;
        if (users != null) {
            for (TestUser user : users) {
                hash = 31 * hash + (user == null ? 13 : user.hashCode());
            }
        }

        return hash;
    }

    /**
     * Gets the users.
     * 
     * @return the users
     */
    public TestUser[] getUsers() {
        if (users == null)
            users = new ArrayList<>();
        return users.toArray(new TestUser[users.size()]);
    }

    /**
     * Sets the users.
     * 
     * @param users
     *            the new users
     */
    public void setUsers( TestUser[] users ) {
        this.users = new ArrayList<>();
        this.users.addAll(Arrays.asList(users));
    }

}
